from dataclasses import dataclass

from .base import Filter
from ..context import ctx
from ..utils import listify


@dataclass
class ChanMemberStatus(Filter):
    value: str | list[str]

    def __call__(self):
        if chat_member := ctx.new_chat_member:
            return chat_member.status in listify(self.value)
        return False



from . import handlers as on
from . import objects as objects
from . import requests as bot
from .base import UserProxyModel, Constants, exceptions as exc, UserModel
from .context import ctx
from .ext import handle
from .loader import logger
from .run import run
from .utils import html



# Copied from aiogram

from __future__ import annotations

import html
from abc import ABC, abstractmethod
from typing import Generator, List, Optional, cast

from ..objects import MessageEntity


class TextDecoration(ABC):
    def apply_entity(self, entity: MessageEntity, text: str) -> str:
        if entity.type in {"bot_command", "url", "mention", "phone_number"}:
            # These entities should not be changed
            return text
        if entity.type in {"bold", "italic", "spoiler", "code", "underline", "strikethrough"}:
            return cast(str, getattr(self, entity.type)(value=text))
        if entity.type == "pre":
            return (
                self.pre_language(value=text, language=entity.language)
                if entity.language
                else self.pre(value=text)
            )
        if entity.type == "text_mention":
            return self.link(value=text, link=f"tg://user?id={entity.user.id}")
        if entity.type == "text_link":
            return self.link(value=text, link=cast(str, entity.url))

        return self.quote(text)

    def unparse(self, text: str, entities: Optional[List[MessageEntity]] = None) -> str:
        return "".join(
            self._unparse_entities(
                self._add_surrogates(text), sorted(entities, key=lambda item: item.offset) if entities else []
            )
        )

    def _unparse_entities(
            self,
            text: bytes,
            entities: List[MessageEntity],
            offset: Optional[int] = None,
            length: Optional[int] = None,
    ) -> Generator[str, None, None]:
        if offset is None:
            offset = 0
        length = length or len(text)

        for index, entity in enumerate(entities):
            if entity.offset * 2 < offset:
                continue
            if entity.offset * 2 > offset:
                yield self.quote(self._remove_surrogates(text[offset: entity.offset * 2]))
            start = entity.offset * 2
            offset = entity.offset * 2 + entity.length * 2

            sub_entities = list(
                filter(lambda e: e.offset * 2 < (offset or 0), entities[index + 1:])
            )
            yield self.apply_entity(
                entity,
                "".join(
                    self._unparse_entities(
                        text, sub_entities, offset=start, length=offset
                    )
                ),
            )

        if offset < length:
            yield self.quote(self._remove_surrogates(text[offset:length]))

    @staticmethod
    def _add_surrogates(text: str):
        return text.encode('utf-16-le')

    @staticmethod
    def _remove_surrogates(text: bytes):
        return text.decode('utf-16-le')

    @abstractmethod
    def link(self, value: str, link: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def bold(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def italic(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def spoiler(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def code(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def pre(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def pre_language(self, value: str, language: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def underline(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def strikethrough(self, value: str) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def quote(self, value: str) -> str:  # pragma: no cover
        pass


class HtmlDecoration(TextDecoration):
    def link(self, value: str, link: str) -> str:
        return f'<a href="{link}">{value}</a>'

    def bold(self, value: str) -> str:
        return f"<b>{value}</b>"

    def italic(self, value: str) -> str:
        return f"<i>{value}</i>"

    def spoiler(self, value: str) -> str:
        return f'<span class="tg-spoiler">{value}</span>'

    def code(self, value: str) -> str:
        return f"<code>{value}</code>"

    def pre(self, value: str) -> str:
        return f"<pre>{value}</pre>"

    def pre_language(self, value: str, language: str) -> str:
        return f'<pre><code class="language-{language}">{value}</code></pre>'

    def underline(self, value: str) -> str:
        return f"<u>{value}</u>"

    def strikethrough(self, value: str) -> str:
        return f"<s>{value}</s>"

    def quote(self, value: str) -> str:
        return html.escape(value, quote=False)


html_decoration = HtmlDecoration()


def entities_to_html(text: str, entities: list[MessageEntity] = None) -> str:
    entities = entities or []
    return html_decoration.unparse(text, entities)



from __future__ import annotations

import typing
from dataclasses import dataclass

import bson
import mongoengine as me

from .base_document import BaseDocument
from .base_model import BaseModel
from ..utils import cast

ModelT = typing.TypeVar('ModelT', bound='UserModel')


def get_model_name(obj: UserModel | type[UserModel]):
    if isinstance(obj, UserModel):
        return obj.__class__.__name__
    return obj.__name__


class UserModelDocument(BaseDocument):
    user_id: int = me.IntField(required=True)
    model: str = me.StringField(required=True)
    committed: bool = me.BooleanField(default=False)
    current: bool = me.BooleanField(default=False)
    data: dict = me.DictField()

    meta = {
        'collection': '__UserModels__',
    }


@dataclass
class UserModel(BaseModel):
    id: bson.ObjectId = None

    # TODO
    # def __init__(self, **attrs):
    #     fields = list(self.__class__.__annotations__) + ['id']
    #
    #     for field in fields:
    #         setattr(self, field, attrs.pop(field, None))
    #
    #     if attrs:
    #         attrs_names = list(attrs)
    #         model_name = get_model_name(self)
    #         raise ValueError(f'Unexpected attrs {attrs_names} for model {model_name}')

    @classmethod
    def make_query(cls, filters: dict) -> dict:
        from ..context import ctx

        filters = {f'data__{key}': value for key, value in filters.items()}
        return {'user_id': ctx.user_id, 'model': get_model_name(cls)} | filters

    @classmethod
    def find(cls: type[ModelT], **filters) -> ModelT | None:
        doc = UserModelDocument.get_doc(**cls.make_query(filters))

        if doc is None:
            return None
        else:
            return cast(doc.data | {'id': doc.id}, cls)

    @classmethod
    def find_all(cls: type[ModelT], **filters) -> list[ModelT]:
        docs = UserModelDocument.get_docs(**cls.make_query(filters))
        return cast([doc.data | {'id': doc.id} for doc in docs], list[cls])

    def save(self: ModelT, as_current: bool = False) -> ModelT:
        from ..context import ctx

        doc = UserModelDocument(user_id=ctx.user_id, model=get_model_name(self))

        if as_current:
            query = {'user_id': ctx.user_id, 'model': get_model_name(self), 'current': True}
            prev_doc = UserModelDocument.get_doc(**query)

            if prev_doc:
                prev_doc.current = False
                prev_doc.save()

            doc.current = True

        doc.data = cast(self, dict)
        doc.id = self.id
        self.id = doc.save().id
        return self

    def delete(self: ModelT) -> None:
        UserModelDocument(id=self.id).delete()

    @classmethod
    def current(cls) -> ModelT | None:
        from ..context import ctx

        query = {'user_id': ctx.user_id, 'model': get_model_name(cls), 'current': True}
        doc = UserModelDocument.get_doc(**query)

        if doc is None:
            return None
        else:
            return cast(doc.data | {'id': doc.id}, cls)
